@@ -260,15 +260,10 @@ index_init(const string & dbpath, const Xapian::Stem & stemmer,
 		// + 1 so that old_lastdocid is a valid subscript.
 		updated.resize(old_lastdocid + 1);
 	    }
-	    try {
-		Xapian::valueno slot = use_ctime ? VALUE_CTIME : VALUE_LASTMOD;
-		string ubound = db.get_value_upper_bound(slot);
-		if (!ubound.empty())
-		    last_altered_max = binary_string_to_int(ubound);
-	    } catch (const Xapian::UnimplementedError &) {
-		numeric_limits<time_t> n;
-		last_altered_max = n.max();
-	    }
+	    Xapian::valueno slot = use_ctime ? VALUE_CTIME : VALUE_LASTMOD;
+	    string ubound = db.get_value_upper_bound(slot);
+	    if (!ubound.empty())
+		last_altered_max = binary_string_to_int(ubound);
 	}
     } else {
 	db = Xapian::WritableDatabase(dbpath, Xapian::DB_CREATE_OR_OVERWRITE);
@@ -954,6 +949,21 @@ index_mimetype(const string & file, const string & urlterm, const string & url,
 				d.get_size(), d.get_mtime());
 		return;
 	    }
+
+	    cmd = "unzip -p";
+	    append_filename_argument(cmd, file);
+	    cmd += " docProps/core.xml";
+	    try {
+		MetaXmlParser metaparser;
+		metaparser.parse(stdout_to_string(cmd, false));
+		title = metaparser.title;
+		keywords = metaparser.keywords;
+		// FIXME: topic = metaparser.topic;
+		sample = metaparser.sample;
+		author = metaparser.author;
+	    } catch (const ReadError&) {
+		// Ignore errors as not all XPS files contain this file.
+	    }
 	} else if (mimetype == "text/csv") {
 	    // Currently we assume that text files are UTF-8 unless they have a
 	    // byte-order mark.
