@@ -4,7 +4,7 @@
 /* Copyright 1999,2000,2001 BrightStation PLC
  * Copyright 2001 Sam Liddicott
  * Copyright 2001,2002 Ananova Ltd
- * Copyright 2002,2003,2004,2005,2006,2007,2008,2009,2010,2011,2014,2015,2017,2018 Olly Betts
+ * Copyright 2002-2022 Olly Betts
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -70,6 +70,23 @@ static bool verbose;
 static int addcount;
 static int repcount;
 static int delcount;
+static int skipcount;
+
+/** What to do if there's a UNIQUE action but a record doesn't use it.
+ */
+static enum {
+    UNIQUE_ERROR,
+    UNIQUE_WARN_NEW,
+    UNIQUE_NEW,
+    UNIQUE_WARN_SKIP,
+    UNIQUE_SKIP
+} unique_missing = UNIQUE_WARN_NEW;
+
+/// Track if UNIQUE action is unused in the current record.
+static bool unique_unused;
+
+/// Track if the current record is being skipping.
+static bool skipping_record = false;
 
 static inline bool
 prefix_needs_colon(const string & prefix, unsigned ch)
@@ -111,7 +128,7 @@ const char * action_names[] = {
 };
 
 // For debugging:
-#define DUMP_ACTION(A) cout << action_names[(A).get_action()] << "(" << (A).get_string_arg() << "," << (A).get_num_arg() << ")" << endl
+#define DUMP_ACTION(A) cout << action_names[(A).get_action()] << "(" << (A).get_string_arg() << "," << (A).get_num_arg() << ")\n"
 
 class Action {
   public:
@@ -213,6 +230,8 @@ squash(string& s, const string& chars)
 
 enum diag_type { DIAG_ERROR, DIAG_WARN, DIAG_NOTE };
 
+static unsigned error_count = 0;
+
 static void
 report_location(enum diag_type type,
 		const string& filename,
@@ -222,14 +241,15 @@ report_location(enum diag_type type,
     cerr << filename;
     if (line != 0) {
 	cerr << ':' << line;
-    }
-    if (pos != string::npos) {
-	// The first column is numbered 1.
-	cerr << ':' << pos + 1;
+	if (pos != string::npos) {
+	    // The first column is numbered 1.
+	    cerr << ':' << pos + 1;
+	}
     }
     switch (type) {
 	case DIAG_ERROR:
 	    cerr << ": error: ";
+	    ++error_count;
 	    break;
 	case DIAG_WARN:
 	    cerr << ": warning: ";
@@ -245,31 +265,46 @@ report_useless_action(const string &file, size_t line, size_t pos,
 		      const string &action)
 {
     report_location(DIAG_WARN, file, line, pos);
-    cerr << "Index action '" << action << "' has no effect" << endl;
+    cerr << "Index action '" << action << "' has no effect\n";
 
     static bool given_left_to_right_warning = false;
     if (!given_left_to_right_warning) {
 	given_left_to_right_warning = true;
 	report_location(DIAG_NOTE, file, line, pos);
-	cerr << "Actions are executed from left to right" << endl;
+	cerr << "Actions are executed from left to right\n";
     }
 }
 
+static bool index_spec_uses_unique = false;
+
 static map<string, vector<Action>> index_spec;
 
+// Like std::getline() but handle \r\n line endings too.
+static istream&
+getline_portable(istream& stream, string& line)
+{
+    istream& result = getline(stream, line);
+    // Trim multiple \r characters, since that seems the best way to handle
+    // that case.
+    line.resize(line.find_last_not_of('\r') + 1);
+    return result;
+}
+
 static void
 parse_index_script(const string &filename)
 {
     ifstream script(filename.c_str());
     if (!script.is_open()) {
 	report_location(DIAG_ERROR, filename);
-	cerr << strerror(errno) << endl;
+	cerr << strerror(errno) << '\n';
 	exit(1);
     }
     string line;
     size_t line_no = 0;
-    // Line number where we saw a `unique` action, or -1 if we haven't.
-    int unique_line_no = -1;
+    // Line number where we saw a `unique` action, or 0 if we haven't.
+    int unique_line_no = 0;
+    // Offset into line unique_line_no where the `unique` action was.
+    size_t unique_pos = 0;
     while (getline(script, line)) {
 	++line_no;
 	vector<string> fields;
@@ -284,10 +319,9 @@ parse_index_script(const string &filename)
 	while (true) {
 	    if (!C_isalnum(*i)) {
 		report_location(DIAG_ERROR, filename, line_no, i - s.begin());
-		cerr << "field name must start with alphanumeric" << endl;
-		exit(1);
+		cerr << "field name must start with alphanumeric\n";
 	    }
-	    j = find_if(i, s.end(),
+	    j = find_if(i + 1, s.end(),
 			[](char ch) { return !C_isalnum(ch) && ch != '_'; });
 	    fields.push_back(string(i, j));
 	    i = find_if(j, s.end(), [](char ch) { return !C_isspace(ch); });
@@ -299,8 +333,10 @@ parse_index_script(const string &filename)
 	    }
 	    if (i == j) {
 		report_location(DIAG_ERROR, filename, line_no, i - s.begin());
-		cerr << "bad character '" << *i << "' in fieldname" << endl;
-		exit(1);
+		cerr << "bad character '" << *i << "' in field name\n";
+		++i;
+		i = find_if(i, s.end(), [](char ch) { return !C_isspace(ch); });
+		if (i == s.end()) break;
 	    }
 	}
 	Xapian::termcount weight = 1;
@@ -408,7 +444,8 @@ parse_index_script(const string &filename)
 			    code = Action::UNHTML;
 			} else if (action == "unique") {
 			    code = Action::UNIQUE;
-			    min_args = max_args = 1;
+			    min_args = 1;
+			    max_args = 2;
 			}
 			break;
 		    case 'v':
@@ -430,15 +467,17 @@ parse_index_script(const string &filename)
 			if (action == "weight") {
 			    code = Action::WEIGHT;
 			    min_args = max_args = 1;
-			    takes_integer_argument = true;
+			    // Don't set takes_integer_argument since we parse
+			    // it with parse_unsigned() and issue an error there
+			    // - setting takes_integer_argument would give a
+			    // double error for arguments with a decimal point.
 			}
 			break;
 		}
 	    }
 	    if (code == Action::BAD) {
 		report_location(DIAG_ERROR, filename, line_no, action_pos);
-		cerr << "Unknown index action '" << action << "'" << endl;
-		exit(1);
+		cerr << "Unknown index action '" << action << "'\n";
 	    }
 	    auto i_after_action = i;
 	    i = find_if(i, s.end(), [](char ch) { return !C_isspace(ch); });
@@ -448,15 +487,14 @@ parse_index_script(const string &filename)
 		    report_location(DIAG_WARN, filename, line_no,
 				    i_after_action - s.begin());
 		    cerr << "putting spaces between the action and '=' is "
-			    "deprecated." << endl;
+			    "deprecated\n";
 		}
 
 		if (max_args == 0) {
 		    report_location(DIAG_ERROR, filename, line_no,
 				    i - s.begin());
 		    cerr << "Index action '" << action
-			 << "' doesn't take an argument" << endl;
-		    exit(1);
+			 << "' doesn't take an argument\n";
 		}
 
 		++i;
@@ -465,7 +503,7 @@ parse_index_script(const string &filename)
 		    report_location(DIAG_WARN, filename, line_no,
 				    i - s.begin());
 		    cerr << "putting spaces between '=' and the argument is "
-			    "deprecated." << endl;
+			    "deprecated\n";
 		}
 
 		vector<string> vals;
@@ -482,8 +520,8 @@ parse_index_script(const string &filename)
 			    if (i == s.end()) {
 				report_location(DIAG_ERROR, filename, line_no,
 						s.size());
-				cerr << "No closing quote" << endl;
-				exit(1);
+				cerr << "No closing quote\n";
+				break;
 			    }
 			    arg.append(j, i);
 			    if (*i++ == '"')
@@ -494,9 +532,9 @@ parse_index_script(const string &filename)
 bad_escaping:
 				report_location(DIAG_ERROR, filename, line_no,
 						i - s.begin());
-				cerr << "Bad escaping in quoted action argument"
-				     << endl;
-				exit(1);
+				cerr << "Bad escaping in quoted action "
+					"argument\n";
+				break;
 			    }
 
 			    char ch = *i;
@@ -520,32 +558,49 @@ bad_escaping:
 				    if (++i == s.end())
 					goto bad_escaping;
 				    char ch1 = *i;
+				    if (!C_isxdigit(ch1)) {
+bad_hex_digit:
+					report_location(DIAG_ERROR, filename,
+							line_no, i - s.begin());
+					cerr << "Bad hex digit in escaping\n";
+					--i;
+					break;
+				    }
 				    if (++i == s.end())
 					goto bad_escaping;
 				    char ch2 = *i;
-				    if (!C_isxdigit(ch1) ||
-					!C_isxdigit(ch2))
-					goto bad_escaping;
+				    if (!C_isxdigit(ch2)) {
+					goto bad_hex_digit;
+				    }
 				    ch = hex_digit(ch1) << 4 |
 					 hex_digit(ch2);
 				    break;
 				}
 				default:
-				    goto bad_escaping;
+				    report_location(DIAG_ERROR, filename,
+						    line_no, i - s.begin());
+				    cerr << "Bad escape sequence '\\" << ch
+					 << "'\n";
+				    break;
 			    }
 			    arg += ch;
 			    j = i + 1;
 			}
 			vals.emplace_back(std::move(arg));
 			if (i == s.end() || C_isspace(*i)) break;
-			if (*i != ',') {
+			if (*i == ',') {
+			    ++i;
+			} else {
 			    report_location(DIAG_ERROR, filename, line_no,
 					    i - s.begin());
 			    cerr << "Unexpected character '" << *i
-				 << "' after closing quote" << endl;
-			    exit(1);
+				 << "' after closing quote\n";
+			    do {
+				++i;
+			    } while (i != s.end() && *i != ',' && !C_isspace(*i));
+			    if (*i != ',') break;
+			    ++i;
 			}
-			++i;
 		    } else if (max_args > 1) {
 			// Unquoted argument, split on comma.
 			i = find_if(j, s.end(),
@@ -567,10 +622,8 @@ bad_escaping:
 		    if (vals.size() == max_args) {
 			report_location(DIAG_ERROR, filename, line_no,
 					i - s.begin());
-			cerr << "Index action '" << action
-			     << "' takes at most " << max_args << " arguments"
-			     << endl;
-			exit(1);
+			cerr << "Index action '" << action << "' takes at most "
+			     << max_args << " arguments\n";
 		    }
 		}
 
@@ -578,15 +631,15 @@ bad_escaping:
 		    report_location(DIAG_ERROR, filename, line_no,
 				    i - s.begin());
 		    if (min_args == max_args) {
-			cerr << "Index action '" << action
-			     << "' requires " << min_args << " arguments"
-			     << endl;
-			exit(1);
+			cerr << "Index action '" << action << "' requires "
+			     << min_args << " arguments\n";
+		    } else {
+			cerr << "Index action '" << action << "' requires "
+				"at least " << min_args << " arguments\n";
 		    }
-		    cerr << "Index action '" << action
-			 << "' requires at least " << min_args << " arguments"
-			 << endl;
-		    exit(1);
+		    // Allow action handling code to assume there are min_args
+		    // arguments.
+		    vals.resize(min_args);
 		}
 
 		string val;
@@ -597,10 +650,10 @@ bad_escaping:
 		if (takes_integer_argument) {
 		    auto dot = val.find('.');
 		    if (dot != string::npos) {
-			report_location(DIAG_WARN, filename, line_no,
+			report_location(DIAG_ERROR, filename, line_no,
 					j - s.begin() + dot);
 			cerr << "Index action '" << action
-			     << "' takes an integer argument" << endl;
+			     << "' takes an integer argument\n";
 		    }
 		}
 		switch (code) {
@@ -608,10 +661,10 @@ bad_escaping:
 			if (val != "unix" &&
 			    val != "unixutc" &&
 			    val != "yyyymmdd") {
-			    report_location(DIAG_ERROR, filename, line_no);
-			    cerr << "Invalid parameter '" << val << "' for "
-				    "action 'date'" << endl;
-			    exit(1);
+			    report_location(DIAG_ERROR, filename, line_no,
+					    j - s.begin());
+			    cerr << "Invalid parameter '" << val
+				 << "' for action 'date'\n";
 			}
 			actions.emplace_back(code, action_pos, val);
 			break;
@@ -624,7 +677,13 @@ bad_escaping:
 			// We don't push an Action for WEIGHT - instead we
 			// store it ready to use in the INDEX and INDEXNOPOS
 			// Actions.
-			weight = atoi(val.c_str());
+			if (!parse_unsigned(val.c_str(), weight)) {
+			    report_location(DIAG_ERROR, filename, line_no,
+					    j - s.begin());
+			    cerr << "Index action 'weight' takes a "
+				    "non-negative integer argument\n";
+			    weight = 0;
+			}
 			if (useless_weight_pos != string::npos) {
 			    report_useless_action(filename, line_no,
 						  useless_weight_pos, action);
@@ -632,17 +691,20 @@ bad_escaping:
 			useless_weight_pos = action_pos;
 			break;
 		    case Action::PARSEDATE: {
-			if (val.find("%Z") != val.npos) {
-			    report_location(DIAG_ERROR, filename, line_no);
-			    cerr << "Parsing timezone names with %Z is not supported" << endl;
-			    exit(1);
+			auto bad_code = val.find("%Z");
+			if (bad_code != val.npos) {
+			    report_location(DIAG_ERROR, filename, line_no,
+					    j - s.begin() + bad_code);
+			    cerr << "Parsing timezone names with %Z is not "
+				    "supported\n";
 			}
 #ifndef HAVE_STRUCT_TM_TM_GMTOFF
-			if (val.find("%z") != val.npos) {
-			    report_location(DIAG_ERROR, filename, line_no);
-			    cerr << "Parsing timezone offsets with %z is not supported on "
-				    "this platform" << endl;
-			    exit(1);
+			bad_code = val.find("%z");
+			if (bad_code != val.npos) {
+			    report_location(DIAG_ERROR, filename, line_no,
+					    j - s.begin() + bad_code);
+			    cerr << "Parsing timezone offsets with %z is not "
+				    "supported on this platform\n";
 			}
 #endif
 			actions.emplace_back(code, action_pos, val);
@@ -650,9 +712,9 @@ bad_escaping:
 		    }
 		    case Action::SPLIT: {
 			if (val.empty()) {
-			    report_location(DIAG_ERROR, filename, line_no);
-			    cerr << "Split delimiter can't be empty" << endl;
-			    exit(1);
+			    report_location(DIAG_ERROR, filename, line_no,
+					    j - s.begin());
+			    cerr << "Split delimiter can't be empty\n";
 			}
 			int operation = Action::SPLIT_NONE;
 			if (vals.size() >= 2) {
@@ -665,10 +727,14 @@ bad_escaping:
 			    } else if (vals[1] == "prefixes") {
 				operation = Action::SPLIT_PREFIXES;
 			    } else {
-				report_location(DIAG_ERROR, filename, line_no);
+				// FIXME: Column should be for where the `op`
+				// parameter starts, which this isn't if the
+				// value is quoted, contains escape sequences,
+				// etc.
+				report_location(DIAG_ERROR, filename, line_no,
+						i - s.begin() - vals[1].size());
 				cerr << "Bad split operation '" << vals[1]
-				     << "'" << endl;
-				exit(1);
+				     << "'\n";
 			    }
 			}
 			actions.emplace_back(code, action_pos, val, operation);
@@ -688,19 +754,36 @@ bad_escaping:
 			actions.emplace_back(code, action_pos, val);
 			break;
 		    case Action::UNIQUE:
-			if (unique_line_no >= 0) {
+			if (unique_line_no) {
 			    report_location(DIAG_ERROR, filename, line_no,
 					    action_pos);
-			    cerr << "Index action 'unique' used more than once"
-				 << endl;
+			    cerr << "Index action 'unique' used more than "
+				    "once\n";
 			    report_location(DIAG_NOTE, filename,
-					    unique_line_no);
-			    cerr << "Previously used here" << endl;
-			    exit(1);
+					    unique_line_no, unique_pos);
+			    cerr << "Previously used here\n";
 			}
 			unique_line_no = line_no;
+			unique_pos = action_pos;
 			if (boolmap.find(val) == boolmap.end())
 			    boolmap[val] = Action::UNIQUE;
+			if (vals.size() >= 2) {
+			    if (vals[1] == "missing=error") {
+				unique_missing = UNIQUE_ERROR;
+			    } else if (vals[1] == "missing=new") {
+				unique_missing = UNIQUE_NEW;
+			    } else if (vals[1] == "missing=warn+new") {
+				unique_missing = UNIQUE_WARN_NEW;
+			    } else if (vals[1] == "missing=skip") {
+				unique_missing = UNIQUE_SKIP;
+			    } else if (vals[1] == "missing=warn+skip") {
+				unique_missing = UNIQUE_WARN_SKIP;
+			    } else {
+				report_location(DIAG_ERROR, filename, line_no);
+				cerr << "Bad unique parameter '" << vals[1]
+				     << "'\n";
+			    }
+			}
 			actions.emplace_back(code, action_pos, val);
 			break;
 		    case Action::GAP: {
@@ -711,8 +794,7 @@ bad_escaping:
 			    report_location(DIAG_ERROR, filename, line_no,
 					    obj.get_pos() + 3 + 1);
 			    cerr << "Index action 'gap' takes a strictly "
-				    "positive integer argument" << endl;
-			    exit(1);
+				    "positive integer argument\n";
 			}
 			break;
 		    }
@@ -724,8 +806,7 @@ bad_escaping:
 			    report_location(DIAG_ERROR, filename, line_no,
 					    obj.get_pos() + 4 + 1);
 			    cerr << "Index action 'hash' takes an integer "
-				    "argument which must be at least 6" << endl;
-			    exit(1);
+				    "argument which must be at least 6\n";
 			}
 			break;
 		    }
@@ -742,7 +823,6 @@ bad_escaping:
 				cerr << "Index action '" << action_names[code]
 				     << "' only support ASCII characters "
 					"currently\n";
-				exit(1);
 			    }
 			}
 			actions.emplace_back(code, action_pos, val);
@@ -760,12 +840,11 @@ bad_escaping:
 				    i_after_action - s.begin());
 		    if (min_args == max_args) {
 			cerr << "Index action '" << action << "' requires "
-			     << min_args << " arguments" << endl;
-			exit(1);
+			     << min_args << " arguments\n";
+		    } else {
+			cerr << "Index action '" << action << "' requires "
+				"at least " << min_args << " arguments\n";
 		    }
-		    cerr << "Index action '" << action << "' requires at least "
-			 << min_args << " arguments" << endl;
-		    exit(1);
 		}
 		switch (code) {
 		    case Action::INDEX:
@@ -829,15 +908,16 @@ bad_escaping:
 	map<string, Action::type>::const_iterator boolpfx;
 	for (boolpfx = boolmap.begin(); boolpfx != boolmap.end(); ++boolpfx) {
 	    if (boolpfx->second == Action::UNIQUE) {
-		report_location(DIAG_WARN, filename, line_no);
+		report_location(DIAG_WARN, filename, unique_line_no,
+				unique_pos);
 		cerr << "Index action 'unique=" << boolpfx->first
-		     << "' without 'boolean=" << boolpfx->first << "'" << endl;
+		     << "' without 'boolean=" << boolpfx->first << "'\n";
 		static bool given_doesnt_imply_boolean_warning = false;
 		if (!given_doesnt_imply_boolean_warning) {
 		    given_doesnt_imply_boolean_warning = true;
-		    report_location(DIAG_NOTE, filename, line_no);
-		    cerr << "'unique' doesn't implicitly add a boolean term"
-			 << endl;
+		    report_location(DIAG_NOTE, filename, unique_line_no,
+				    unique_pos);
+		    cerr << "'unique' doesn't implicitly add a boolean term\n";
 		}
 	    }
 	}
@@ -862,9 +942,14 @@ bad_escaping:
 
     if (index_spec.empty()) {
 	report_location(DIAG_ERROR, filename, line_no);
-	cerr << "No rules found in index script" << endl;
+	cerr << "No rules found in index script\n";
+    }
+
+    if (error_count) {
 	exit(1);
     }
+
+    index_spec_uses_unique = (unique_line_no > 0);
 }
 
 static bool
@@ -937,26 +1022,24 @@ run_actions(vector<Action>::const_iterator action_it,
 		size_t len = value.length();
 		if (len & 1) {
 		    report_location(DIAG_ERROR, fname, line_no);
-		    cerr << "hextobin: input must have even length"
-			 << endl;
-		} else {
-		    string output;
-		    output.reserve(len / 2);
-		    for (size_t j = 0; j < len; j += 2) {
-			char a = value[j];
-			char b = value[j + 1];
-			if (!C_isxdigit(a) || !C_isxdigit(b)) {
-			    report_location(DIAG_ERROR, fname, line_no);
-			    cerr << "hextobin: input must be all hex "
-				    "digits" << endl;
-			    goto badhex;
-			}
-			char r = (hex_digit(a) << 4) | hex_digit(b);
-			output.push_back(r);
-		    }
-		    value = std::move(output);
+		    cerr << "hextobin: input must have even length\n";
+		    exit(1);
 		}
-badhex:
+
+		string output;
+		output.reserve(len / 2);
+		for (size_t j = 0; j < len; j += 2) {
+		    char a = value[j];
+		    char b = value[j + 1];
+		    if (!C_isxdigit(a) || !C_isxdigit(b)) {
+			report_location(DIAG_ERROR, fname, line_no);
+			cerr << "hextobin: input must be all hex digits\n";
+			exit(1);
+		    }
+		    char r = (hex_digit(a) << 4) | hex_digit(b);
+		    output.push_back(r);
+		}
+		value = std::move(output);
 		break;
 	    }
 	    case Action::LOWER:
@@ -979,7 +1062,7 @@ badhex:
 		// If there's no input, just issue a warning.
 		if (value.empty()) {
 		    report_location(DIAG_WARN, fname, line_no);
-		    cerr << "Empty filename in LOAD action" << endl;
+		    cerr << "Empty filename in LOAD action\n";
 		    break;
 		}
 		bool truncated = false;
@@ -989,9 +1072,8 @@ badhex:
 			       value, truncated)) {
 		    report_location(DIAG_ERROR, fname, line_no);
 		    cerr << "Couldn't load file '" << filename << "': "
-			 << strerror(errno) << endl;
-		    value.resize(0);
-		    break;
+			 << strerror(errno) << '\n';
+		    exit(1);
 		}
 		if (!truncated) break;
 	    }
@@ -1139,11 +1221,32 @@ badhex:
 		break;
 	    }
 	    case Action::UNIQUE: {
-		// If there's no text, just issue a warning.
+		unique_unused = false;
+
 		if (value.empty()) {
-		    report_location(DIAG_WARN, fname, line_no);
-		    cerr << "Ignoring UNIQUE action on empty text"
-			 << endl;
+		    enum diag_type diag = DIAG_WARN;
+		    switch (unique_missing) {
+		      case UNIQUE_ERROR:
+			diag = DIAG_ERROR;
+			/* FALLTHRU */
+		      case UNIQUE_WARN_NEW:
+		      case UNIQUE_WARN_SKIP:
+			report_location(diag, fname, line_no);
+			cerr << "UNIQUE action on empty text\n";
+		      default:
+			break;
+		    }
+		    switch (unique_missing) {
+		      case UNIQUE_ERROR:
+			exit(1);
+		      case UNIQUE_SKIP:
+		      case UNIQUE_WARN_SKIP:
+			skipping_record = true;
+			break;
+		      case UNIQUE_NEW:
+		      case UNIQUE_WARN_NEW:
+			break;
+		    }
 		    break;
 		}
 
@@ -1178,7 +1281,7 @@ badhex:
 		if (*end) {
 		    report_location(DIAG_WARN, fname, line_no);
 		    cerr << "Trailing characters in VALUENUMERIC: '"
-			 << value << "'" << endl;
+			 << value << "'\n";
 		}
 		doc.add_value(action.get_num_arg(),
 			      Xapian::sortable_serialise(dbl));
@@ -1203,11 +1306,10 @@ badhex:
 		    report_location(DIAG_WARN, fname, line_no);
 		    cerr << "valuepacked \"" << value << "\" ";
 		    if (errno == ERANGE) {
-			cerr << "out of range";
+			cerr << "out of range\n";
 		    } else {
-			cerr << "not an unsigned integer";
+			cerr << "not an unsigned integer\n";
 		    }
-		    cerr << endl;
 		}
 		int valueslot = action.get_num_arg();
 		doc.add_value(valueslot, int_to_binary_string(word));
@@ -1224,7 +1326,7 @@ badhex:
 		    if (!parse_signed(value.c_str(), t)) {
 			report_location(DIAG_WARN, fname, line_no);
 			cerr << "Date value (in secs) for action DATE "
-				"must be an integer - ignoring" << endl;
+				"must be an integer - ignoring\n";
 			break;
 		    }
 		    struct tm *tm = localtime(&t);
@@ -1236,7 +1338,7 @@ badhex:
 		    if (!parse_signed(value.c_str(), t)) {
 			report_location(DIAG_WARN, fname, line_no);
 			cerr << "Date value (in secs) for action DATE "
-				"must be an integer - ignoring" << endl;
+				"must be an integer - ignoring\n";
 			break;
 		    }
 		    struct tm *tm = gmtime(&t);
@@ -1247,7 +1349,7 @@ badhex:
 		    if (value.length() != 8) {
 			report_location(DIAG_WARN, fname, line_no);
 			cerr << "date=yyyymmdd expects an 8 character value "
-				"- ignoring" << endl;
+				"- ignoring\n";
 			break;
 		    }
 		    yyyymmdd = value;
@@ -1271,7 +1373,7 @@ badhex:
 		if (ret == NULL) {
 		    report_location(DIAG_WARN, fname, line_no);
 		    cerr << "\"" << value << "\" doesn't match format "
-			    "\"" << dateformat << '\"' << endl;
+			    "\"" << dateformat << '\"' << '\n';
 		    break;
 		}
 
@@ -1280,7 +1382,7 @@ badhex:
 		    cerr << "\"" << value << "\" not fully matched by "
 			    "format \"" << dateformat << "\" "
 			    "(\"" << ret << "\" left over) but "
-			    "indexing anyway" << endl;
+			    "indexing anyway\n";
 		}
 #ifdef HAVE_STRUCT_TM_TM_GMTOFF
 		auto gmtoff = tm.tm_gmtoff;
@@ -1307,43 +1409,39 @@ index_file(const char *fname, istream &stream,
 {
     string line;
     size_t line_no = 0;
-    while (!stream.eof() && getline(stream, line)) {
+    while (!stream.eof() && getline_portable(stream, line)) {
 	++line_no;
+	// Allow blank lines before the first record and multiple blank lines
+	// between records.
+	if (line.empty()) continue;
+
 	Xapian::Document doc;
 	indexer.set_document(doc);
 	Xapian::docid docid = 0;
 	map<string, list<string>> fields;
 	bool seen_content = false;
+	skipping_record = false;
+	unique_unused = index_spec_uses_unique;
 	while (!line.empty()) {
-	    // Cope with files from MS Windows (\r\n end of lines).
-	    // Trim multiple \r characters, since that seems the best way
-	    // to handle that case.
-	    string::size_type last = line.find_last_not_of('\r');
-	    if (last == string::npos) break;
-	    line.resize(last + 1);
-
 	    string::size_type eq = line.find('=');
 	    if (eq == string::npos && !line.empty()) {
-		report_location(DIAG_ERROR, fname, line_no, line.size());
-		cerr << "expected = somewhere in this line" << endl;
-		// FIXME: die or what?
+		report_location(DIAG_ERROR, fname, line_no);
+		cerr << "Expected = somewhere in this line\n";
+		exit(1);
 	    }
 	    string field(line, 0, eq);
 	    string value(line, eq + 1, string::npos);
-	    while (getline(stream, line)) {
+	    line.clear();
+	    while (getline_portable(stream, line)) {
 		++line_no;
 		if (line.empty() || line[0] != '=') break;
-		// Cope with files from MS Windows (\r\n end of lines).
-		// Trim multiple \r characters, since that seems the best way
-		// to handle that case.
-		last = line.find_last_not_of('\r');
-		// line[0] == '=', so last != string::npos.
-		// Replace the '=' with a '\n' so we don't have to use substr.
+		// Replace the '=' with a '\n'.
 		line[0] = '\n';
-		line.resize(last + 1);
 		value += line;
 	    }
 
+	    if (skipping_record) continue;
+
 	    // Default to not indexing spellings.
 	    indexer.set_flags(Xapian::TermGenerator::flags(0));
 
@@ -1355,15 +1453,42 @@ index_file(const char *fname, istream &stream,
 			field, fname, line_no,
 			docid);
 	    if (this_field_is_content) seen_content = true;
-	    if (stream.eof()) break;
 	}
 
-	// If we haven't seen any fields (other than unique identifiers)
-	// the document is to be deleted.
-	if (!seen_content) {
+	if (unique_unused) {
+	    enum diag_type diag = DIAG_WARN;
+	    switch (unique_missing) {
+	      case UNIQUE_ERROR:
+		diag = DIAG_ERROR;
+		/* FALLTHRU */
+	      case UNIQUE_WARN_NEW:
+	      case UNIQUE_WARN_SKIP:
+		report_location(diag, fname, line_no);
+		cerr << "UNIQUE action unused in this record\n";
+	      default:
+		break;
+	    }
+	    switch (unique_missing) {
+	      case UNIQUE_ERROR:
+		exit(1);
+	      case UNIQUE_SKIP:
+	      case UNIQUE_WARN_SKIP:
+		skipping_record = true;
+		break;
+	      case UNIQUE_NEW:
+	      case UNIQUE_WARN_NEW:
+		break;
+	    }
+	}
+
+	if (skipping_record) {
+	    ++skipcount;
+	} else if (!seen_content) {
+	    // We haven't seen any fields (other than unique identifiers)
+	    // so the document is to be deleted.
 	    if (docid) {
 		database.delete_document(docid);
-		if (verbose) cout << "Del: " << docid << endl;
+		if (verbose) cout << "Del: " << docid << '\n';
 		++delcount;
 	    }
 	} else {
@@ -1383,11 +1508,11 @@ index_file(const char *fname, istream &stream,
 	    // Add the document to the database
 	    if (docid) {
 		database.replace_document(docid, doc);
-		if (verbose) cout << "Replace: " << docid << endl;
+		if (verbose) cout << "Replace: " << docid << '\n';
 		++repcount;
 	    } else {
 		docid = database.add_document(doc);
-		if (verbose) cout << "Add: " << docid << endl;
+		if (verbose) cout << "Add: " << docid << '\n';
 		++addcount;
 	    }
 	}
@@ -1395,7 +1520,7 @@ index_file(const char *fname, istream &stream,
 
     // Commit after each file to make sure all changes from that file make it
     // in.
-    if (verbose) cout << "Committing: " << endl;
+    if (verbose) cout << "Committing\n";
     database.commit();
 }
 
@@ -1468,7 +1593,7 @@ try {
 		} catch (const Xapian::InvalidArgumentError &) {
 		    cerr << "Unknown stemming language '" << optarg << "'.\n";
 		    cerr << "Available language names are: "
-			 << Xapian::Stem::get_available_languages() << endl;
+			 << Xapian::Stem::get_available_languages() << '\n';
 		    return 1;
 		}
 		break;
@@ -1496,6 +1621,7 @@ try {
     addcount = 0;
     repcount = 0;
     delcount = 0;
+    skipcount = 0;
 
     if (argc == 2) {
 	// Read from stdin.
@@ -1507,20 +1633,23 @@ try {
 	    if (stream) {
 		index_file(argv[i], stream, database, indexer);
 	    } else {
-		cerr << "Can't open file " << argv[i] << endl;
+		cerr << "Can't open file " << argv[i] << '\n';
 	    }
 	}
     }
 
-    cout << "records (added, replaced, deleted) = (" << addcount << ", "
-	 << repcount << ", " << delcount << ")" << endl;
+    cout << "records (added, replaced, deleted, skipped) = ("
+	 << addcount << ", "
+	 << repcount << ", "
+	 << delcount << ", "
+	 << skipcount << ")\n";
 } catch (const Xapian::Error &error) {
-    cerr << "Exception: " << error.get_description() << endl;
+    cerr << "Exception: " << error.get_description() << '\n';
     exit(1);
 } catch (const std::bad_alloc &) {
-    cerr << "Exception: std::bad_alloc" << endl;
+    cerr << "Exception: std::bad_alloc\n";
     exit(1);
 } catch (...) {
-    cerr << "Unknown Exception" << endl;
+    cerr << "Unknown Exception\n";
     exit(1);
 }
